from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models.subreddit import Subreddit


class BanhammerException(Exception):
    def __init__(self, msg: str):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class NotModerator(BanhammerException):
    def __init__(self, user: str, sub: 'Subreddit'):
        super().__init__(f"/u/{user} doesn't moderate /r/{sub}.")


class NoRedditInstance(BanhammerException):
    def __init__(self):
        super().__init__("No <apraw.Reddit> instance was given to the <banhammer.Subreddit> object.")


class NoItemGiven(BanhammerException):
    def __init__(self):
        super().__init__("No <banhammer.RedditItem> was given to the <banhammer.Reaction> instance.")


class NotEligibleItem(BanhammerException):
    def __init__(self):
        super().__init__("The <banhammer.RedditItem> given to the <banhammer.Reaction> object cannot be handled.")



from typing import TYPE_CHECKING, Union

import apraw
import discord
from apraw.models import (Comment, Message, ModAction, ModmailConversation,
                          ModmailMessage, Submission, Subreddit)

from ..const import logger

if TYPE_CHECKING:
    from .subreddit import Subreddit


class RedditItem:

    def __init__(self, item: Union[Comment, ModmailConversation, ModmailMessage,
                                   Submission], subreddit: 'Subreddit', source: str):
        self.item = item
        self.id = item.id
        self.type = "submission" if isinstance(
            item, Submission) else "comment" if isinstance(
            item, Comment) else "modmail" if type(item) in [
            ModmailMessage, ModmailConversation] else "mod action"
        self.subreddit = subreddit
        self.source = source
        self._author = None

    def __repr__(self):
        return f"<{self.type.title()} id={self.id} source={self.source}>"

    async def get_message(self):
        return await self.subreddit.banhammer.message_builder.get_item_message(self)

    async def to_dict(self):
        return {
            "url": self.url,
            "author": await self.get_author_name(),
            "type": self.type,
            "source": self.source,
            "body": self.body
        }

    async def get_embed(self, *args, **kwargs):
        return await self.subreddit.banhammer.message_builder.get_item_embed(self, *args, **kwargs)

    def format_reply(self, reply: str):
        return self.subreddit.banhammer.message_builder.format_reply(self, reply)

    def get_ban_message(self, ban_duration: int):
        return self.subreddit.banhammer.message_builder.get_ban_message(self, ban_duration)

    async def get_author(self):
        if not self._author:
            if isinstance(self.item, ModAction):
                try:
                    self._author = await self.item.mod()
                except Exception as e:
                    logger.error(f"Failed to retrieve item author in {self.item}: {e}")
            elif not isinstance(self.item, ModmailConversation):
                try:
                    self._author = await self.item.author()
                except Exception as e:
                    logger.error(f"Failed to retrieve item author in {self.item}: {e}")
            else:
                self._author = self.item.authors[0]
        return self._author

    async def is_author_removed(self):
        author = await self.get_author()
        author_removed = not author
        if author and not isinstance(author, dict):
            author_removed = not hasattr(author, "name")
        elif author:
            author_removed = author.get("isDeleted", False)
        return author_removed

    async def get_author_name(self):
        if await self.is_author_removed():
            return "[deleted]"
        else:
            author = await self.get_author()
            if not isinstance(author, dict):
                return author.name
            else:
                return author["name"]

    @property
    def reactions(self):
        return self.subreddit.get_reactions(self.item)

    async def add_reactions(self, message: discord.Message):
        for r in self.reactions:
            try:
                await message.add_reaction(r.emoji)
            except Exception as e:
                logger.error(f"Failed to add reaction '{r.emoji}': {e}")
                continue

    def get_reaction(self, emoji: str):
        return self.subreddit.get_reaction(emoji, self.item)

    @property
    def url(self):
        return get_item_url(self.item)

    @property
    def body(self):
        if self.type == "submission":
            return self.item.selftext[:1021] + (self.item.selftext[1021:] and "...") or "Empty"
        elif self.type == "comment":
            return self.item.body[:1021] + (self.item.body[1021:] and "...")
        elif self.type == "modmail":
            return self.item.body_md[:1021] + (self.item.body_md[1021:] and "...")
        elif self.type == "mod action":
            return self.item.action


def get_item_url(item: RedditItem):
    if isinstance(item, Submission):
        return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item.id}"
    elif isinstance(item, Comment):
        return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item.link_id.replace('t3_', '')}/_/{item.id}"
    elif isinstance(item, ModmailConversation):
        return "https://mod.reddit.com/mail/all/" + item.id
    elif isinstance(item, ModmailMessage):
        return "https://mod.reddit.com/mail/all/" + item.conversation.id
    elif isinstance(item, Message):
        if item.was_comment:
            return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item.link_id.replace('t3_', '')}/_/{item.id}"
        else:
            return "https://www.reddit.com/message/messages/{}" + str(item)
    elif isinstance(item, Subreddit):
        return "https://www.reddit.com/r/" + item.display_name
    return ""



import asyncio
import enum
from typing import TYPE_CHECKING, Any, Awaitable, Callable, List, Union

if TYPE_CHECKING:
    from .item import RedditItem
    from .subreddit import Subreddit


class ItemAttribute(enum.Enum):
    SUBREDDIT = "subreddit"
    AUTHOR = "author"
    MOD = "mod"


class GeneratorIdentifier(enum.Enum):
    NEW = "new"
    COMMENTS = "comments"
    REPORTS = "reports"
    MAIL = "mail"
    QUEUE = "queue"
    MOD_ACTIONS = "mod_actions"

    def __str__(self):
        return str(self.value)


class EventFilter:

    def __init__(self, attribute: ItemAttribute, *args, **kwargs):
        self._attribute = attribute
        self._values = args
        self._reverse = kwargs.get("reverse", False)

    async def is_item_valid(self, item: 'RedditItem'):
        if not self._values:
            return True

        match = True

        if self._attribute == ItemAttribute.MOD:
            if item.type != "mod action":
                return False
            mod_name = await item.get_author_name()
            match = any(mod_name.lower() == str(v).lower() for v in self._values)
        elif self._attribute == ItemAttribute.AUTHOR:
            author_name = await item.get_author_name()
            match = any(author_name.lower() == str(v).lower() for v in self._values)
        elif self._attribute == ItemAttribute.SUBREDDIT:
            match = any(str(item.subreddit).lower() == str(v).lower() for v in self._values)

        return (not self._reverse and match) or (self._reverse and not match)

    def is_subreddit_valid(self, subreddit: 'Subreddit'):
        return self._attribute != ItemAttribute.SUBREDDIT or any(
            str(subreddit).lower() == str(v).lower() for v in self._values) or not self._values


class EventHandler:

    def __init__(self, callback: Callable[['RedditItem'], Awaitable[None]], identifier: GeneratorIdentifier, *args):
        self._callback = callback
        self._identifiers = {identifier}
        self._filters = list(args)
        self._takes_self = False

    @classmethod
    def new(cls, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            return cls.create_event_handler(func, GeneratorIdentifier.NEW, **kwargs)
        return wrapper

    @classmethod
    def comments(cls, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            return cls.create_event_handler(func, GeneratorIdentifier.COMMENTS, **kwargs)
        return wrapper

    @classmethod
    def mail(cls, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            return cls.create_event_handler(func, GeneratorIdentifier.MAIL, **kwargs)
        return wrapper

    @classmethod
    def queue(cls, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            return cls.create_event_handler(func, GeneratorIdentifier.QUEUE, **kwargs)
        return wrapper

    @classmethod
    def reports(cls, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            return cls.create_event_handler(func, GeneratorIdentifier.REPORTS, **kwargs)
        return wrapper

    @classmethod
    def mod_actions(cls, *args, **kwargs):
        def wrapper(func: Callable[['RedditItem'], Awaitable[None]]):
            event_filter = EventFilter(ItemAttribute.MOD, *kwargs.get("mods", tuple()), *args)
            create_args = (event_filter,) if event_filter._values else tuple()
            return cls.create_event_handler(func, GeneratorIdentifier.MOD_ACTIONS, *create_args, **kwargs)
        return wrapper

    @classmethod
    def create_event_handler(cls, handler: Union['EventHandler', Callable[['RedditItem'], Awaitable[None]]],
                             identifier: GeneratorIdentifier, *args, **kwargs):
        if not isinstance(handler, EventHandler) and not asyncio.iscoroutinefunction(handler):
            raise TypeError("Event handler must be a coroutine function or of type <banhammer.models.EventHandler>.")

        args = (*args, *getattr(handler, "_filters", tuple()))

        values = tuple(*kwargs.get("subreddits", tuple()), str(kwargs.get("subreddit", "")))
        if values:
            event_filter = EventFilter(ItemAttribute.SUBREDDIT, values)
            args = (*args, event_filter)

        if not isinstance(handler, EventHandler):
            handler = EventHandler(handler, identifier, *args)
        else:
            handler._identifiers.add(identifier)
            handler._filters.extend(args)

        return handler

    @classmethod
    def filter(cls, attribute: ItemAttribute, *args, **kwargs):
        def wrapper(handler: Union['EventHandler', Callable[['RedditItem'], Awaitable[None]]]):
            if not isinstance(handler, EventHandler) and not asyncio.iscoroutinefunction(handler):
                raise TypeError("Event handler must be a coroutine function or of type <banhammer.models.EventHandler>.")

            handler._filters = getattr(handler, "_filters", list())
            handler._filters.append(EventFilter(attribute, *args, **kwargs))
            return handler
        return wrapper

    async def __call__(self, identifier: GeneratorIdentifier, *args):
        valid = identifier in self._identifiers

        if not valid:
            return

        for f in self._filters:
            if not await f.is_item_valid(args[-1]):
                valid = False
                break

        if valid:
            await self._callback(*args)

    def get_sub_funcs(self, subreddits: List['Subreddit']):
        for subreddit in subreddits:
            if all(f.is_subreddit_valid(subreddit) for f in self._filters if f._attribute == ItemAttribute.SUBREDDIT):
                for identifier in self._identifiers:
                    yield getattr(subreddit, f"get_{identifier}"), identifier



from .event_handler import (EventFilter, EventHandler, GeneratorIdentifier,
                            ItemAttribute)
from .item import RedditItem
from .message_builder import MessageBuilder
from .reaction import (Reaction, ReactionHandler, ReactionPayload,
                       get_reactions, ignore_reactions)
from .subreddit import Subreddit



import re
from urllib.parse import urlparse

import apraw

from ..const import logger
from ..models import RedditItem

REDDIT_URL_PATTERN = re.compile(
    r"((https:\/\/)?((www|old|np|mod)\.)?(reddit|redd){1}(\.com|\.it){1}([a-zA-Z0-9\/_]+))")
POST_URL_PATTERN = re.compile(
    r"/r(?:/(?P<subreddit>\w+))/comments(?:/(?P<submission>\w+))(?:/\w+/(?P<comment>\w+))?")
MODMAIL_URL_PATTERN = re.compile(
    r"https://mod.reddit.com/mail/all/(?P<id>\w+)")


async def get_item(reddit: apraw.Reddit, subreddits, str):
    for u in REDDIT_URL_PATTERN.findall(str):
        if is_url(u[0]):
            item = await get_item_from_url(reddit, subreddits, u[0])
            if item:
                return item
            else:
                continue
    return None


async def get_item_from_url(reddit: apraw.Reddit, subreddits, url):
    match = MODMAIL_URL_PATTERN.search(url)
    if match:
        for subreddit in subreddits:
            try:
                modmail = await subreddit._subreddit.modmail(match.group("id"))
                if hasattr(modmail, "subject"):
                    return RedditItem(modmail, subreddit, "url")
            except Exception as e:
                logger.error(f"Failed to fetch modmail by ID '{match.group('id')}': {e}")
        return None

    match = POST_URL_PATTERN.search(url)
    if match and not match.group("comment"):
        try:
            item = await reddit.submission(match.group("submission"))
        except Exception as e:
            logger.error(f"Failed to fetch submission by ID '{match.group('submission')}': {e}")
            return None
    elif match:
        try:
            item = await reddit.comment(match.group("comment"))
        except Exception as e:
            logger.error(f"Failed to fetch comment by ID '{match.group('comment')}': {e}")
            return None
    else:
        return None

    try:
        item_subreddit = await item.subreddit()
    except Exception as e:
        logger.error(f"Failed to retrieve item {item} subreddit: {e}")
        return None
    else:
        for sub in subreddits:
            s = await sub.get_subreddit()
            if s.id == item_subreddit.id:
                return RedditItem(item, sub, "url")
    return None


def is_url(url):
    check = urlparse(url)
    return check.scheme != "" and check.netloc != ""



from datetime import datetime
from typing import List

import discord
from apraw.models import ModmailConversation, ModmailMessage
from discord.utils import escape_markdown

from ..const import BANHAMMER_PURPLE, BOT_DISCLAIMER, logger
from .item import RedditItem
from .reaction import ReactionPayload
from .subreddit import Subreddit


class MessageBuilder:

    async def get_item_message(self, item: RedditItem):
        if item.type in ["submission", "comment"]:
            author_name = escape_markdown(await item.get_author_name())
            subreddit = escape_markdown(str(item.subreddit))
            return f"New {item.type} on /r/{subreddit} by /u/{author_name}!\n\n" + \
                f"{item.url}\n\n" + \
                f"**Title:** {item.item.title}\n**Body:**\n{item.body}"
        elif item.type == "modmail":
            author_name = escape_markdown(await item.get_author_name())
            subreddit = escape_markdown(str(item.subreddit))
            subject = escape_markdown(item.item.conversation.subject)
            return f"New message in modmail conversation '{subject}' on /r/{subreddit} by /u/{author_name}!" + \
                f"\n\n{item.body}"
        else:
            author_name = escape_markdown(await item.get_author_name())
            subreddit = escape_markdown(str(item.subreddit))
            return f"New action taken by /u/{author_name} on /r/{subreddit}: `{item.body}`"

    async def get_item_embed(self, item: RedditItem, embed_color: discord.Color = None,
                             embed_template: discord.Embed = None):
        embed = embed_template or discord.Embed(
            colour=embed_color or item.subreddit.banhammer.embed_color
        )

        if isinstance(item.item, ModmailConversation):
            try:
                embed.timestamp = datetime.utcfromtimestamp(item.item.last_updated)
            except Exception as e:
                logger.error(f"Error setting timestamp <{item.item.last_updated}>: {e}")
        elif isinstance(item.item, ModmailMessage):
            try:
                timestamp = int(float(item.item.conversation.last_updated))
                embed.timestamp = datetime.utcfromtimestamp(timestamp)
            except Exception as e:
                logger.error(f"Error converting timestamp <{item.item.conversation.last_updated}> to int: {e}")
        else:
            try:
                embed.timestamp = datetime.utcfromtimestamp(item.item._data["created_utc"])
            except Exception as e:
                logger.error(f"Error setting timestamp <{item.item._data['created_utc']}>: {e}")

        if item.type in ["submission", "comment"]:
            if item.source == "reports":
                title = f"{item.type.title()} reported on /r/{item.subreddit} by /u/{await item.get_author_name()}!"
            else:
                title = f"New {item.type} on /r/{item.subreddit} by /u/{await item.get_author_name()}!"
        elif item.type == "modmail":
            title = f"New message in modmail conversation '{item.item.conversation.subject}' on /r/{item.subreddit} by /u/{await item.get_author_name()}!"
        else:
            title = f"New action taken by /u/{await item.get_author_name()} on /r/{item.subreddit}!"

        subreddit = await item.subreddit.get_subreddit()
        embed.set_author(name=title,
                         url=item.url or discord.Embed.Empty,
                         icon_url=subreddit.community_icon or discord.Embed.Empty)

        if item.type == "submission":
            embed.add_field(name="Title", value=escape_markdown(item.item.title), inline=False)

            if item.item.link_flair_text:
                embed.description = f"Flair: `{item.item.link_flair_text}`"

            if item.item.is_self:
                embed.add_field(name="Body", value=item.body, inline=False)
            elif "i.redd.it" in item.item.url or any(item.item.url.endswith(ext) for ext in (".jpg", ".jpeg", ".png", ".gif")):
                embed.set_image(url=item.item.url)
            elif not item.item._data.get("poll_data", None):
                embed.add_field(name="URL", value=escape_markdown(item.item.url), inline=False)

            if item.item._data.get("poll_data", None):
                options = [
                    f"◽ {escape_markdown(option['text'])}" for option in item.item.poll_data["options"]]
                embed.add_field(name="Poll", value="\n".join(options), inline=False)
            elif item.item._data.get("media_metadata", None):
                for _, media in item.item.media_metadata.items():
                    if media["e"] == "Image":
                        added = False
                        if isinstance(media.get("s", None), dict) and "u" in media["s"]:
                            embed.set_image(url=media["s"]["u"])
                            added = True
                        elif isinstance(media.get("p", None), list):
                            try:
                                embed.set_image(url=media["p"][-1]["u"])
                                added = True
                            except BaseException:
                                logger.warning(f"Unexpected media metadata: {media}")
                        else:
                            logger.warning(f"Unexpected media metadata: {media}")

                        if added:
                            break

            if item.source == "reports":
                user_reports = "\n".join(f"{r[1]} {r[0]}" for r in item.item.user_reports)
                if user_reports:
                    embed.add_field(name="User Reports", value=user_reports, inline=False)

                mod_reports = "\n".join(f"{r[1]} {r[0]}" for r in item.item.mod_reports)
                if mod_reports:
                    embed.add_field(name="Mod Reports", value=mod_reports, inline=False)
        elif item.type in ("comment", "modmail"):
            embed.description = item.body
        elif item.type == "mod action":
            embed.description = f"Action: `{item.body}`"

        return embed

    def get_ban_message(self, item: RedditItem, ban_duration: int):
        ban_type = "permanent" if not ban_duration else "temporary"
        disclaimer = BOT_DISCLAIMER.format(item.subreddit.contact_url)
        return f"Our moderator team has reviewed [this post]({item.url}) and decided to give you a {ban_type} ban. " \
               f"If you wish to appeal this ban, please respond to this message.\n\n{disclaimer}"

    def format_reply(self, item: RedditItem, reply: str):
        disclaimer = BOT_DISCLAIMER.format(item.subreddit.contact_url)
        return f"{reply}\n\n{disclaimer}"

    async def get_payload_message(self, payload: ReactionPayload):
        if not payload.actions:
            payload.actions.append("dismissed")

        user = escape_markdown(payload.user)
        author_name = escape_markdown(await payload.item.get_author_name())
        url = escape_markdown(payload.item.url)

        return f"**{payload.item.type.title()} {' and '.join(payload.actions)} by {user}!**\n\n" \
               f"{payload.item.type.title()} by /u/{author_name}:\n\n{url}"

    async def get_payload_embed(self, payload: ReactionPayload, embed_color: discord.Color = None,
                                embed_template: discord.Embed = None):
        if not payload.actions:
            payload.actions.append("dismissed")

        embed = embed_template or discord.Embed(
            colour=embed_color or payload.item.subreddit.banhammer.embed_color
        )

        author_name = escape_markdown(await payload.item.get_author_name())

        embed.description = f"[{payload.item.type.title()}]({payload.item.url}) by /u/{author_name}."
        embed.timestamp = datetime.utcnow()

        embed.set_author(name=f"{payload.item.type.title()} {' and '.join(payload.actions)} by {payload.user}!")

        return embed

    def get_reactions_embed(self, subreddits: List[Subreddit], embed_color: discord.Color = None,
                            embed_template: discord.Embed = None):
        """
        Load an embed with all the configured reactions per subreddit.

        Parameters
        ----------
        subreddits : List[Subreddit]
            The subreddits to be included in the overview.
        embed_color : discord.Color
            The color to be used for the embed, if not specified, the
            :attr:`~banhammer.Banhammer.embed_color` is used.

        Returns
        -------
        embed: discord.Embed
            The embed listing all the configured reactions per subreddit.
        """
        embed = embed_template or discord.Embed(
            colour=embed_color or subreddits[0].banhammer.embed_color if subreddits else BANHAMMER_PURPLE
        )

        embed.title = "Configured reactions"
        embed.timestamp = datetime.utcnow()

        for sub in subreddits:
            embed.add_field(name="/r/" + str(sub),
                            value="\n".join([repr(r) for r in sub.reactions]),
                            inline=False)
        return embed

    def get_subreddits_embed(self, subreddits: List[Subreddit], embed_color: discord.Color = None,
                             embed_template: discord.Embed = None):
        """
        Load an embed with all the configured subreddits and their enabled streams.

        Parameters
        ----------
        subreddits : List[Subreddit]
            The subreddits to be included in the overview.
        embed_color : discord.Color
            The color to be used for the embed, if not specified, the
            :attr:`~banhammer.Banhammer.embed_color` is used.

        Returns
        -------
        embed: discord.Embed
            The embed of all the subreddits and their enabled streams.
        """
        embed = embed_template or discord.Embed(
            colour=embed_color or subreddits[0].banhammer.embed_color if subreddits else BANHAMMER_PURPLE
        )

        embed.title = "Subreddits' statuses"
        embed.description = "\n".join([s.status for s in subreddits])
        embed.timestamp = datetime.utcnow()

        return embed

    async def get_subreddit_reactions_embed(self, subreddit: Subreddit, embed_color: discord.Color = None,
                                            embed_template: discord.Embed = None):
        embed = embed_template or discord.Embed(
            colour=embed_color or subreddit.banhammer.embed_color
        )

        embed.timestamp = datetime.utcnow()

        sub = await subreddit.get_subreddit()
        embed.set_author(
            name=f"/r/{subreddit} Configured Reactions",
            url=f"https://www.reddit.com/r/{subreddit}/wiki/banhammer-reactions",
            icon_url=sub.community_icon or discord.Embed.Empty)

        fields = list()
        for reaction in subreddit.reactions:
            name = ' '.join(w.title() if w.islower() else w for w in reaction.__comments__.splitlines()
                            [0].replace('#', '', 1).strip().split()) if reaction.__comments__ else ""
            title = f"{name}: {reaction}" if name else f"`{reaction}`"
            text = repr(reaction).replace(str(reaction) + " | ", "")

            if reaction.reply:
                text.replace(" | reply", "")
                text += f"\n\n**Reply**\n>>> {reaction.reply.splitlines()[0]}"
                fields.append({"name": title, "value": text})
            else:
                fields = [{"name": title, "value": text}, *fields]
        for field in fields:
            embed.add_field(**field)

        return embed



import os
from typing import TYPE_CHECKING, List

from apraw.utils import BoundedSet

from ..const import logger
from .item import RedditItem
from .reaction import get_reactions

if TYPE_CHECKING:
    from ..banhammer import Banhammer


class Subreddit:

    def __init__(self, bh: 'Banhammer', **opts):
        self.banhammer = bh
        self.reddit = bh.reddit

        self.name = opts["subreddit"] if "subreddit" in opts else ""
        self.name = self.name.replace("r/", "").replace("/", "")

        self.stream_new = opts.get("stream_new", True)
        self.stream_comments = opts.get("stream_comments", False)
        self.stream_reports = opts.get("stream_reports", True)
        self.stream_mail = opts.get("stream_mail", True)
        self.stream_queue = opts.get("stream_queue", True)
        self.stream_mod_actions = opts.get("stream_mod_actions", True)

        self._new_ids = BoundedSet(301)
        self._comment_ids = BoundedSet(301)
        self._report_ids = BoundedSet(301)
        self._mail_ids = BoundedSet(301)
        self._queue_ids = BoundedSet(301)
        self._mod_action_ids = BoundedSet(301)

        self._skip_new = True
        self._skip_comments = True
        self._skip_reports = True
        self._skip_mail = True
        self._skip_queue = True
        self._skip_mod_actions = True

        self._subreddit = None

        self.custom_emotes = opts.get("custom_emotes", True)

        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "reactions.yaml"))
        with open(path, encoding="utf8") as f:
            content = f.read()
            self.reactions = get_reactions(content)["reactions"]

    def __str__(self):
        return self.name

    @property
    def status(self):
        str = "/r/" + self.name

        if self.stream_new:
            str += " | New Posts"
        if self.stream_comments:
            str += " | Comments"
        if self.stream_reports:
            str += " | Reports"
        if self.stream_mail:
            str += " | Mod-Mail"
        if self.stream_queue:
            str += " | Mod-Queue"

        return str

    @property
    def contact_url(self):
        return "https://www.reddit.com/message/compose/?to=/r/" + self.name

    async def get_subreddit(self):
        if not self._subreddit:
            self._subreddit = await self.reddit.subreddit(self.name)
        return self._subreddit

    async def setup(self):
        subreddit = await self.get_subreddit()
        settings = await subreddit.mod.settings()
        self.stream_new = settings.spam_links != "all" and settings.spam_selfposts != "all"
        self.stream_comments = settings.spam_comments == "all"
        self.stream_queue = settings.spam_links == "all" or settings.spam_selfposts == "all"

    async def load_reactions(self):
        subreddit = await self.get_subreddit()
        loaded = False

        if self.custom_emotes:
            try:
                reaction_page = await subreddit.wiki.page("banhammer-reactions")
                reacts = get_reactions(reaction_page.content_md)["reactions"]
                if reacts:
                    self.reactions = reacts
                    loaded = True
            except Exception as e:
                logger.error(f"Couldn't load wikipage: {e}")

        if not loaded:
            path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "reactions.yaml"))
            with open(path, encoding="utf8") as f:
                try:
                    await subreddit.wiki.create("banhammer-reactions", f.read(), "Reactions not found")
                except Exception as e:
                    logger.error(f"Couldn't create wikipage: {e}")

    async def get_reactions_embed(self, *args, **kwargs):
        return await self.banhammer.message_builder.get_subreddit_reactions_embed(self, *args, **kwargs)

    def get_reactions(self, item: RedditItem):
        return [r for r in self.reactions if r.eligible(item)]

    def get_reaction(self, emoji: str, item: RedditItem):
        for reaction in self.get_reactions(item):
            if reaction.emoji == emoji:
                return reaction

    async def get_new(self):
        subreddit = await self.get_subreddit()
        submissions = [s async for s in subreddit.new()]
        for submission in reversed(submissions):
            if not submission:
                continue
            if submission.id in self._new_ids:
                continue

            self._new_ids.add(submission.id)

            if not self._skip_new:
                item = RedditItem(submission, self, "new")
                yield item

        self._skip_new = False

    async def get_comments(self):
        subreddit = await self.get_subreddit()
        comments = [s async for s in subreddit.comments()]
        for comment in reversed(comments):
            if not comment:
                continue
            if comment.id in self._comment_ids:
                continue

            self._comment_ids.add(comment.id)

            if not self._skip_comments:
                item = RedditItem(comment, self, "new")
                yield item

        self._skip_comments = False

    async def get_reports(self):
        subreddit = await self.get_subreddit()
        items = [s async for s in subreddit.mod.reports()]
        for item in reversed(items):
            if not item:
                continue
            if item.id in self._report_ids:
                continue

            self._report_ids.add(item.id)

            if not self._skip_reports:
                item = RedditItem(item, self, "reports")
                yield item

        self._skip_reports = False

    async def get_mail(self):
        subreddit = await self.get_subreddit()
        conversations = [s async for s in subreddit.modmail.conversations()]
        for conversation in reversed(conversations):
            async for message in conversation.messages():
                if not message:
                    continue
                if message.id in self._mail_ids:
                    continue

                self._mail_ids.add(message.id)

                if not self._skip_mail:
                    message = RedditItem(message, self, "modmail")
                    yield message

        self._skip_mail = False

    async def get_queue(self):
        subreddit = await self.get_subreddit()
        items = [s async for s in subreddit.mod.modqueue()]
        for item in reversed(items):
            if not item:
                continue
            if item.id in self._queue_ids:
                continue

            self._queue_ids.add(item.id)

            if not self._skip_queue:
                item = RedditItem(item, self, "queue")
                yield item

        self._skip_queue = False

    async def get_mod_actions(self, mods: List[str] = list()):
        subreddit = await self.get_subreddit()
        mods = [m.lower() for m in mods]
        actions = [s async for s in subreddit.mod.log()]
        for action in reversed(actions):
            if not action:
                continue
            if action.id in self._mod_action_ids:
                continue
            if action._data["mod"].lower() not in mods and mods:
                continue

            self._mod_action_ids.add(action.id)

            if not self._skip_mod_actions:
                action = RedditItem(action, self, "log")
                yield action

        self._skip_mod_actions = False



import asyncio
import json
import os
import re
from typing import Awaitable, Callable, Dict, Tuple, Union, Any

import apraw
import discord
from apraw.utils import ExponentialCounter

from .const import BANHAMMER_PURPLE, logger
from .models import (EventFilter, EventHandler, GeneratorIdentifier,
                     ItemAttribute, MessageBuilder, ReactionHandler,
                     RedditItem, Subreddit)
from .utils import reddit_helper


class BanhammerMeta(type):

    def __new__(cls, name, bases, dct):
        banhammer = super().__new__(cls, name, bases, dct)

        event_handlers = list()

        for base in reversed(banhammer.__mro__):
            for name, attr in list(base.__dict__.items()):
                if isinstance(attr, EventHandler):
                    attr._takes_self = True
                    event_handlers.append(attr)
                    delattr(base, name)

        banhammer._event_handlers = event_handlers

        return banhammer


class Banhammer(metaclass=BanhammerMeta):
    """
    The main Banhammer class that manages the event loop to poll Reddit and forward items to configured callables.
    """

    def __init__(self, reddit: apraw.Reddit, counter: ExponentialCounter = ExponentialCounter(16), bot: discord.Client = None, embed_color: discord.Colour = BANHAMMER_PURPLE,
                 change_presence: bool = False, message_builder: MessageBuilder = MessageBuilder(), reaction_handler: ReactionHandler = ReactionHandler()):
        """
        Create a Banhammer instance.

        Parameters
        ----------
        reddit : apraw.Reddit
            The Reddit instance with which Subreddits are constructed and requests are made.
        counter : ExponentialCounter, optional
            The counter to use to increase/decrease wait times based on whether items were found.
        bot : discord.Client, optional
            The discord.Client in case presence should be changed, by default ``None``.
        embed_color : discord.Colour, optional
            The color to use for generated embeds, by default rgb(207, 206, 255).
        change_presence : bool, optional
            Whether the client's presence should be changed to "Watching Reddit" when polls are being performed, by default False
        message_builder : MessageBuilder, optional
            An instance of :class:`~banhammer.models.MessageBuilder` which is used by items to generate their embeds and such,
            by default MessageBuilder().
        reaction_handler : ReactionHandler, optional
            An instance of :class:`~banhammer.models.ReactionHandler` which handles reactions and feeds
            :class:`~banhammer.models.ReactionPayload`, by default ReactionHandler().
        """
        self.reddit = reddit
        self.message_builder = message_builder
        self.reaction_handler = reaction_handler
        self.embed_color = embed_color

        self._bot = bot
        self._change_presence = change_presence
        self._counter = counter

        self.subreddits = list()
        self.loop = getattr(self, "loop", None) or asyncio.get_event_loop()

        self._event_handlers = getattr(self, "_event_handlers", list())

    async def add_subreddits(self, *subs):
        """
        Add subreddits for Banhammer to poll in its loop.

        If the given subreddit isn't of type :class:`~banhammer.models.Subreddit`, Banhammer will
        convert it and load the reactions / add default ones to the wiki using the
        :meth:`~banhammer.models.Subreddit.load_reactions` method.

        Parameters
        ----------
        *subs : Tuple[Subreddit]
            The *args subreddits to be added.
        """
        for sub in subs:
            if not isinstance(sub, Subreddit):
                sub = Subreddit(self, subreddit=str(sub))
                await sub.load_reactions()
            self.subreddits.append(sub)

    def remove_subreddit(self, subreddit: Union[Subreddit, apraw.models.Subreddit, str]):
        """
        Remove a subreddit from Banhammer's list.

        Parameters
        ----------
        subreddit : Union[Subreddit, apraw.models.Subreddit, str]
            The subreddit or the name of a subreddit to be removed.

        Returns
        -------
        removed : bool
            Whether a subreddit was found and removed.
        """
        subreddit = str(subreddit).lower().replace("r/", "").replace("/", "")
        for sub in self.subreddits:
            sub = str(sub).lower().replace("r/", "").replace("/", "")
            if sub == subreddit:
                self.subreddits.remove(sub)
                return True
        return False

    def new(self, **kwargs):
        """
        A decorator to subscribe to new posts with the given function.

        The decorator can be used like this:

        .. code-block:: python3

            @bh.new()
            async def handle_new(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_new_handler(func, **kwargs)
        return wrapper

    def add_new_handler(self, func: Callable[[RedditItem], Awaitable[None]], **kwargs):
        """
        Add a function to handle new posts.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when new posts are made.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        return self.add_event_handler(func, GeneratorIdentifier.NEW, **kwargs)

    async def handle_new(self, item: RedditItem):
        """
        Handle Reddit items from the new queues.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.NEW, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_new: {e}")
                await self.on_handler_error(e)

    def comments(self, **kwargs):
        """
        A decorator to subscribe to comments with the given function.

        The decorator can be used like this:

        .. code-block:: python3

            @bh.comments()
            async def handle_comments(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_comments_handler(func, **kwargs)
        return wrapper

    def add_comments_handler(self, func: Callable[[RedditItem], Awaitable[None]], **kwargs):
        """
        Add a function to handle comments.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when comments are made.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        return self.add_event_handler(func, GeneratorIdentifier.COMMENTS, **kwargs)

    async def handle_comments(self, item: RedditItem):
        """
        Handle Reddit items from the comment queues.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.COMMENTS, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_comments: {e}")
                await self.on_handler_error(e)

    def mail(self, **kwargs):
        """
        A decorator to subscribe to modmail with the given function.

        The decorator can be used like this:

        .. code-block:: python3

            @bh.mail()
            async def handle_mail(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_mail_handler(func, **kwargs)
        return wrapper

    def add_mail_handler(self, func: Callable[[RedditItem], Awaitable[None]], **kwargs):
        """
        Add a function to handle modmail.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when modmail are made.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        return self.add_event_handler(func, GeneratorIdentifier.MAIL, **kwargs)

    async def handle_mail(self, item: RedditItem):
        """
        Handle Reddit items from the mail queues.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.MAIL, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_mail: {e}")
                await self.on_handler_error(e)

    def queue(self, **kwargs):
        """
        A decorator to subscribe to unmoderated posts with the given function.

        The decorator can be used like this:

        .. code-block:: python3

            @bh.queue()
            async def handle_queue(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_queue_handler(func, **kwargs)
        return wrapper

    def add_queue_handler(self, func: Callable[[RedditItem], Awaitable[None]], **kwargs):
        """
        Add a function to handle unmoderated posts.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when unmoderated posts are made.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        return self.add_event_handler(func, GeneratorIdentifier.QUEUE, **kwargs)

    async def handle_queue(self, item: RedditItem):
        """
        Handle Reddit items from the modqueue.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.QUEUE, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_queue: {e}")
                await self.on_handler_error(e)

    def reports(self, **kwargs):
        """
        A decorator to subscribe to reported posts with the given function.

        The decorator can be used like this:

        .. code-block:: python3

            @bh.reports()
            async def handle_reports(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_reports_handler(func, **kwargs)
        return wrapper

    def add_reports_handler(self, func: Callable[[RedditItem], Awaitable[None]], **kwargs):
        """
        Add a function to handle reported posts.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when reported posts are made.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        return self.add_event_handler(func, GeneratorIdentifier.REPORTS, **kwargs)

    async def handle_reports(self, item: RedditItem):
        """
        Handle Reddit items from the reports queue.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.REPORTS, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_reports: {e}")
                await self.on_handler_error(e)

    def mod_actions(self, *args, **kwargs):
        """
        A decorator to subscribe to mod actions with the given function.

        The decorator can be used like this, arguments passed will be
        used as the name of moderators to search for if given:

        .. code-block:: python3

            @bh.mod_actions("Anti-Evil Operations")
            async def handle_mod_actions(item: RedditItem):
                pass
        """
        def wrapper(func: Callable[[RedditItem], Awaitable[None]]):
            return self.add_mod_actions_handler(func, *args, **kwargs)
        return wrapper

    def add_mod_actions_handler(self, func: Callable[[RedditItem], Awaitable[None]], *args, **kwargs):
        """
        Add a function to handle mod actions.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when mod actions are made.
        mods : Tuple[str]
            Moderators to search for if specified.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        event_filter = EventFilter(ItemAttribute.MOD, *kwargs.get("mods", tuple()), *args)
        create_args = (event_filter,) if event_filter._values else tuple()
        return self.add_event_handler(func, GeneratorIdentifier.MOD_ACTIONS, *create_args, **kwargs)

    async def handle_mod_actions(self, item: RedditItem):
        """
        Handle Reddit items from the mod action queues.

        The default implementation of this method calls the appropriate handlers that were
        assigned using Banhammer's decorators. If overriden, it is recommended to make a
        ``super()`` call, otherwise handlers may be missed.

        Parameters
        ----------
        item : RedditItem
            The item to forward to handlers.
        """
        for handler in self._event_handlers:
            handler_args = (self, item) if handler._takes_self else (item,)
            try:
                await handler(GeneratorIdentifier.MOD_ACTIONS, *handler_args)
            except Exception as e:
                logger.error(f"Handler error in {handler} with item from handle_mod_actions: {e}")
                await self.on_handler_error(e)

    def add_event_handler(self, func: Callable[[RedditItem], Awaitable[None]],
                          identifier: GeneratorIdentifier, *args, **kwargs):
        """
        Add a function to handle the items from a specific generator in the
        :class:`~banhammer.models.Subreddit` class.

        Parameters
        ----------
        func : Callable[[RedditItem], Awaitable[None]]
            The function that will be called when items are retrieved.
        identifier : GeneratorIdentifier
            The identifier used to retrieve the subreddit's generator as well as the internal handler.
        subreddit : str or Subreddit
            The subreddit to poll with this function.
        """
        event_handler = EventHandler.create_event_handler(func, identifier, *args, **kwargs)
        self._event_handlers.append(event_handler)
        return event_handler

    async def on_handler_error(self, error: Any):
        pass

    async def send_items(self):
        """
        Start the loop to poll Reddit and send items.
        """
        while True:
            if self._bot and self._change_presence:
                try:
                    watching = discord.Activity(type=discord.ActivityType.watching, name="Reddit")
                    await self._bot.change_presence(activity=watching)
                except Exception as e:
                    logger.error(f"Failed to change bot presence: {e}")

            funcs = set()

            if self._event_handlers:
                for handler in self._event_handlers:
                    funcs.update(func for func in handler.get_sub_funcs(self.subreddits))
            else:
                for subreddit in self.subreddits:
                    funcs.update((getattr(subreddit, f"get_{identifier}"), identifier)
                                 for identifier in GeneratorIdentifier)

            found = False
            for func in funcs:
                try:
                    async for item in func[0]():
                        found = True
                        try:
                            await getattr(self, f"handle_{func[1]}")(item)
                        except Exception as e:
                            logger.error(f"Handler error in handle_{func[1]} with item from {func[0]}: {e}")
                            await self.on_handler_error(e)
                except Exception as e:
                    logger.error(f"Error fetching item from {func[0]}: {e}")

            if self._bot is not None and self._change_presence:
                try:
                    await self._bot.change_presence(activity=None)
                except Exception as e:
                    logger.error(f"Failed to change bot presence: {e}")

            wait_time = (found and self._counter.reset()) or self._counter.count()

            await asyncio.sleep(wait_time)

    async def get_item(self, c: Union[str, discord.Embed]):
        """
        Retrieve a :class:`banhammer.models.RedditItem` from a message or embed.

        Parameters
        ----------
        c : Union[str, discord.Embed]
            The message contents or embed to parse.

        Returns
        -------
        item: RedditItem
            The item found by its URL in the message or embed.
        """
        s = str(c) if not isinstance(c, discord.Embed) else json.dumps(c.to_dict())
        return await reddit_helper.get_item(self.reddit, self.subreddits, s)

    def get_reactions_embed(self, *args, **kwargs):
        """
        Load an embed with all the configured reactions per subreddit.

        Parameters
        ----------
        embed_color : discord.Color
            The color to be used for the embed, if not specified, the
            :attr:`~banhammer.Banhammer.embed_color` is used.

        Returns
        -------
        embed: discord.Embed
            The embed listing all the configured reactions per subreddit.
        """
        return self.message_builder.get_reactions_embed(self.subreddits, *args, **kwargs)

    def get_subreddits_embed(self, *args, **kwargs):
        """
        Load an embed with all the configured subreddits and their enabled streams.

        Parameters
        ----------
        embed_color : discord.Color
            The color to be used for the embed, if not specified, the
            :attr:`~banhammer.Banhammer.embed_color` is used.

        Returns
        -------
        embed: discord.Embed
            The embed of all the subreddits and their enabled streams.
        """
        return self.message_builder.get_subreddits_embed(self.subreddits, *args, **kwargs)

    def start(self):
        """
        Add the Banhammer poll loop to the ``asyncio`` event loop.
        """
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "WELCOME.md"))
        with open(path) as f:
            print("")
            BOLD = '\033[1m'
            END = '\033[0m'
            print(re.sub(r"\*\*(.+)\*\*", r"{}\1{}".format(BOLD, END), f.read()))
            print("")

        self.loop.create_task(self.send_items())
